import inspect
from django.conf import settings
from django.apps import apps
from django.contrib.auth import get_user_model
from instance_selector.selectors import (
    ModelAdminInstanceSelector,
    WagtailUserInstanceSelector,
)

from instance_selector.exceptions import ModelAdminLookupFailed

__all__ = ("Registry", "registry")


class Registry:
    def __init__(self):
        self._models = {
            # ('app_label', 'model_name') => Model
        }
        self._selectors = {
            # Model => InstanceSelector instance
        }

    def get_model(self, app_label, model_name):
        key = (app_label, model_name)
        if key not in self._models:
            model = apps.get_model(app_label=app_label, model_name=model_name)
            self.register_model(app_label, model_name, model)
        return self._models[key]

    def get_instance_selector(self, model):
        User = get_user_model()
        if model not in self._selectors:
            from wagtail.admin.menu import admin_menu, settings_menu

            model_admin = self._find_model_admin_in_menu(admin_menu, model)
            if not model_admin:
                model_admin = self._find_model_admin_in_menu(settings_menu, model)

            if model_admin:
                instance_selector = ModelAdminInstanceSelector(model_admin=model_admin)
                self.register_instance_selector(model, instance_selector)
            elif model is User:
                if "wagtail.users" in settings.INSTALLED_APPS:
                    # Wagtail uses bespoke functional views for Users, so we apply a preconfigured
                    # variant at the last opportunity. This reduces setup difficulty, while preserving
                    # the ability to apply overrides downstream
                    self.register_instance_selector(
                        model, WagtailUserInstanceSelector()
                    )
            else:
                raise ModelAdminLookupFailed(
                    "Cannot find model admin for %s. You may need to register a model with wagtail's admin or "
                    "register an instance selector" % model
                )

        return self._selectors[model]

    def register_instance_selector(self, model, instance_selector):
        if model in self._selectors:
            raise Exception(
                "%s has already been registered. Cannot register %s -> %s"
                % (model, model, instance_selector)
            )

        if inspect.isclass(instance_selector):
            raise Exception(
                "Expected an instance of a class, but received %s. You may need to call your class before "
                "registering it" % instance_selector
            )

        self._selectors[model] = instance_selector

    def register_model(self, app_label, model_name, model):
        key = (app_label, model_name)
        if key in self._models:
            raise Exception(
                "%s has already been registered to %s. Cannot register %s -> %s"
                % (key, self._models[key], key, model)
            )
        self._models[key] = model

    def clear(self):
        """
        Forces the registry to re-discover all models and selectors
        """
        self._models = {}
        self._selectors = {}

    def _find_model_admin_in_menu(self, menu, model):
        for item in menu.registered_menu_items:
            if hasattr(item, "model_admin"):
                model_admin = item.model_admin
                if model_admin.model == model:
                    return model_admin
            if hasattr(item, "menu"):
                model_admin = self._find_model_admin_in_menu(item.menu, model)
                if model_admin:
                    return model_admin


registry = Registry()



from django.utils.functional import cached_property, lazy

from instance_selector.widgets import InstanceSelectorWidget
from instance_selector.registry import registry

from wagtail.blocks import ChooserBlock
from wagtail.coreutils import resolve_model_string
from wagtail.telepath import register
from wagtail.blocks.field_block import FieldBlockAdapter



class InstanceSelectorBlock(ChooserBlock):
    class Meta:
        icon = "placeholder"

    def __init__(self, target_model, **kwargs):
        super().__init__(**kwargs)

        self._target_model = target_model

        if self.meta.icon == "placeholder":
            # The models/selectors may not have been registered yet, depending upon
            # import orders and things, so get the icon lazily
            self.meta.icon = lazy(self.get_instance_selector_icon, str)

    @cached_property
    def target_model(self):
        return resolve_model_string(self._target_model)

    @cached_property
    def widget(self):
        return InstanceSelectorWidget(self.target_model)

    def get_form_state(self, value):
        return self.widget.get_value_data(value)

    def get_instance_selector_icon(self):
        instance_selector = registry.get_instance_selector(self.target_model)
        return instance_selector.get_widget_icon()

    def deconstruct(self):
        name, args, kwargs = super().deconstruct()

        if args:
            args = args[1:]  # Remove the args target_model

        kwargs["target_model"] = self.target_model._meta.label_lower
        return name, args, kwargs

    def value_from_form(self, value):
        # When validation fails, the value might be "" (Wagtail 5.0+)
        # and None if succeeds.
        if value == "":
            value = None

        return super().value_from_form(value)

class InstanceSelectorBlockAdapter(FieldBlockAdapter):
    def js_args(self, block):
        name, widget, meta = super().js_args(block)

        # Fix up the 'icon' item in meta so that it's a string that we can serialize,
        # rather than a lazy reference
        if callable(meta["icon"]):
            meta["icon"] = meta["icon"]()

        return [name, widget, meta]


register(InstanceSelectorBlockAdapter(), InstanceSelectorBlock)



from django.urls import reverse
from django_webtest import WebTest
from django.contrib.auth import get_user_model

from instance_selector.constants import OBJECT_PK_PARAM
from instance_selector.registry import registry
from instance_selector.selectors import (
    BaseInstanceSelector,
    ModelAdminInstanceSelector,
    WagtailUserInstanceSelector,
)
from .test_project.test_app.models import TestModelA, TestModelB, TestModelC
from .test_project.test_app.wagtail_hooks import (
    TestModelAAdmin,
    TestModelBAdmin,
)

User = get_user_model()


class Tests(WebTest):
    """
    Commented out tests are failing for the Wagtail 4.0 + releases.
    
    Im not sure how relevant they are now that the widgets are being rendered
    by javascript.
    """
    def setUp(self):
        TestModelA.objects.all().delete()
        TestModelB.objects.all().delete()
        TestModelC.objects.all().delete()
        self.superuser = User.objects.create_superuser(
            "superuser", "superuser@example.com", "test"
        )
        registry.clear()

    def test_registry_automatically_discovers_models(self):
        self.assertEqual(registry.get_model("test_app", "TestModelB"), TestModelB)

    def test_registry_throws_for_unknown_models(self):
        self.assertRaises(
            LookupError, lambda: registry.get_model("test_app", "SomeUnknownModel")
        )

    def test_registry_automatically_creates_instance_selectors(self):
        selector = registry.get_instance_selector(TestModelB)
        self.assertIsInstance(selector.model_admin, TestModelBAdmin)

    def test_custom_model_can_be_registered(self):
        class Model:
            pass

        registry.register_model("foo", "bar", Model)
        self.assertEqual(registry.get_model("foo", "bar"), Model)

    def test_custom_instance_selector_can_be_registered(self):
        class TestInstanceSelector(BaseInstanceSelector):
            pass

        registry.register_instance_selector(TestModelB, TestInstanceSelector())
        self.assertIsInstance(
            registry.get_instance_selector(TestModelB), TestInstanceSelector
        )

    def test_user_instance_selector_is_automatically_created(self):
        selector = registry.get_instance_selector(User)
        self.assertIsInstance(selector, WagtailUserInstanceSelector)

    def test_widget_renders_during_model_creation(self):
        res = self.app.get("/admin/test_app/testmodelb/create/", user=self.superuser)
        self.assertIn('/static/instance_selector/instance_selector.css', res.text)
        self.assertIn('/static/instance_selector/instance_selector_embed.js', res.text)
        self.assertIn('/static/instance_selector/instance_selector_widget.js', res.text)
        # leaving these commented out for now, as the widget is now rendered by javascript
        # and they may be useful to decide how relevant they are now.
        # self.assertIn('class="instance-selector-widget ', res.text)
        # self.assertIn("create_instance_selector_widget({", res.text)

    def test_widget_renders_during_model_edit_without_value(self):
        b = TestModelB.objects.create()
        res = self.app.get(
            "/admin/test_app/testmodelb/edit/%s/" % b.pk, user=self.superuser
        )
        self.assertIn('/static/instance_selector/instance_selector.css', res.text)
        self.assertIn('/static/instance_selector/instance_selector_embed.js', res.text)
        self.assertIn('/static/instance_selector/instance_selector_widget.js', res.text)
        # leaving these commented out for now, as the widget is now rendered by javascript
        # and they may be useful to decide how relevant they are now.
        # self.assertIn('class="instance-selector-widget ', res.text)
        # self.assertIn("create_instance_selector_widget({", res.text)

    def test_widget_renders_during_model_edit_with_value(self):
        a = TestModelA.objects.create()
        b = TestModelB.objects.create(test_model_a=a)
        res = self.app.get(
            "/admin/test_app/testmodelb/edit/%s/" % b.pk, user=self.superuser
        )
        self.assertIn(
            '<input type="hidden" name="test_model_a" value="%s" id="id_test_model_a">'
            % a.pk,
            res.text,
        )
        self.assertIn(
            '<span class="instance-selector-widget__display__title">TestModelA object (%s)</span>'
            % a.pk,
            res.text,
        )

    def test_widget_can_render_custom_display_data(self):
        class TestInstanceSelector(BaseInstanceSelector):
            def get_instance_display_title(self, instance):
                return "test display title"

            def get_instance_display_image_url(self, instance):
                return "test display image url"

            def get_instance_edit_url(self, instance):
                return "test edit url"

        registry.register_instance_selector(TestModelA, TestInstanceSelector())

        a = TestModelA.objects.create()
        b = TestModelB.objects.create(test_model_a=a)

        res = self.app.get(
            "/admin/test_app/testmodelb/edit/%s/" % b.pk, user=self.superuser
        )
        self.assertIn(
            '/static/instance_selector/instance_selector.css',
            res.text,
        )
        self.assertIn(
            '/static/instance_selector/instance_selector_embed.js',
            res.text,
        )
        self.assertIn(
            '/static/instance_selector/instance_selector_widget.js',
            res.text,
        )
        # leaving these commented out for now, as the widget is now rendered by javascript
        # and they may be useful to decide how relevant they are now.
        # self.assertIn(
        #     '<span class="instance-selector-widget__display__title">test display title</span>',
        #     res.text,
        # )
        # self.assertIn('src="test display image url"', res.text)
        # self.assertIn('href="test edit url"', res.text)

    def test_widget_can_render_custom_display_markup(self):
        class TestInstanceSelector(ModelAdminInstanceSelector):
            def get_instance_display_markup(self, instance):
                return "test display markup"

        registry.register_instance_selector(
            TestModelA, TestInstanceSelector(model_admin=TestModelAAdmin())
        )

        a = TestModelA.objects.create()
        b = TestModelB.objects.create(test_model_a=a)

        res = self.app.get(
            "/admin/test_app/testmodelb/edit/%s/" % b.pk, user=self.superuser
        )
        self.assertIn(
            '/static/instance_selector/instance_selector.css',
            res.text,
        )
        self.assertIn(
            '/static/instance_selector/instance_selector_embed.js',
            res.text,
        )
        self.assertIn(
            '/static/instance_selector/instance_selector_widget.js',
            res.text,
        )
        # leaving this commented out for now, as the widget is now rendered by javascript
        # and they may be useful to decide how relevant they are now.
        # self.assertIn("test display markup", res.text)

    def test_widget_can_lookup_updated_display_information(self):
        selector = registry.get_instance_selector(TestModelA)

        a = TestModelA.objects.create()

        lookup_url = "%s?%s=%s" % (
            reverse(
                "wagtail_instance_selector_lookup",
                kwargs={"app_label": "test_app", "model_name": "TestModelA"},
            ),
            OBJECT_PK_PARAM,
            a.pk,
        )
        res = self.app.get(lookup_url, user=self.superuser)
        self.assertEqual(
            res.json,
            {
                "display_markup": selector.get_instance_display_markup(a),
                "edit_url": "/admin/test_app/testmodela/edit/%s/" % a.pk,
                "pk": "%s" % a.pk,
            },
        )

    def test_widget_can_lookup_updated_custom_display_information(self):
        class TestInstanceSelector(BaseInstanceSelector):
            def get_instance_display_markup(self, instance):
                return "test display markup"

            def get_instance_edit_url(self, instance):
                return "test edit url"

        registry.register_instance_selector(TestModelA, TestInstanceSelector())

        a = TestModelA.objects.create()

        lookup_url = "%s?%s=%s" % (
            reverse(
                "wagtail_instance_selector_lookup",
                kwargs={"app_label": "test_app", "model_name": "TestModelA"},
            ),
            OBJECT_PK_PARAM,
            a.pk,
        )
        res = self.app.get(lookup_url, user=self.superuser)
        self.assertEqual(
            res.json,
            {
                "display_markup": "test display markup",
                "edit_url": "test edit url",
                "pk": "%s" % a.pk,
            },
        )

    def test_widget_lookup_without_pk_will_fail(self):
        lookup_url = reverse(
            "wagtail_instance_selector_lookup",
            kwargs={"app_label": "test_app", "model_name": "TestModelA"},
        )
        res = self.app.get(lookup_url, user=self.superuser, expect_errors=True)
        self.assertEqual(res.status_code, 400)

    def test_embed_view_renders_selector_url(self):
        selector = registry.get_instance_selector(TestModelA)
        embed_url = reverse(
            "wagtail_instance_selector_embed",
            kwargs={"app_label": "test_app", "model_name": "TestModelA"},
        )
        res = self.app.get(embed_url, user=self.superuser)
        self.assertIn(
            '<iframe onload="removeSidebar(this);" src="%s#instance_selector_embed_id:test_app-testmodela-'
            % selector.get_instance_selector_url(),
            res.text,
        )

    def test_embed_view_renders_custom_selector_url(self):
        class TestInstanceSelector(ModelAdminInstanceSelector):
            def get_instance_selector_url(self):
                return "test selector url"

        registry.register_instance_selector(
            TestModelA, TestInstanceSelector(model_admin=TestModelAAdmin())
        )
        embed_url = reverse(
            "wagtail_instance_selector_embed",
            kwargs={"app_label": "test_app", "model_name": "TestModelA"},
        )
        res = self.app.get(embed_url, user=self.superuser)
        self.assertIn(
            '<iframe onload="removeSidebar(this);" src="test selector url#instance_selector_embed_id:test_app-testmodela-',
            res.text,
        )

    def test_blocks_can_render_widget_code(self):
        c = TestModelC.objects.create()
        res = self.app.get(
            "/admin/test_app/testmodelc/edit/%s/" % c.pk, user=self.superuser
        )
        # rendered widget output is embedded in JSON within the data-block attribute
        self.assertIn(
            "class=\&quot;instance-selector-widget instance-selector-widget--unselected instance-selector-widget--required\&quot;",
            res.text,
        )



from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

from instance_selector.widgets import InstanceSelectorWidget

from wagtail.admin.panels import FieldPanel


class InstanceSelectorPanel(FieldPanel):
    model = None
    field_name = None

    def get_form_options(self):
        opts = super().get_form_options()

        # Use the instance selector widget for this option
        opts["widgets"] = {self.field_name: InstanceSelectorWidget(model=self.target_model)}

        return opts

    @property
    def target_model(self):
        return self.db_field.remote_field.model

    def render_as_field(self):
        instance_obj = self.get_chosen_item()
        return mark_safe(
            render_to_string(
                self.field_template,
                {"field": self.bound_field, "instance": instance_obj},
            )
        )



import json
from django.forms import widgets
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from instance_selector.constants import OBJECT_PK_PARAM
from instance_selector.registry import registry

from wagtail.telepath import register
from wagtail.utils.widgets import WidgetWithScript
from wagtail.widget_adapters import WidgetAdapter


class InstanceSelectorWidget(WidgetWithScript, widgets.Input):
    # when looping over form fields, this one should appear in visible_fields, not hidden_fields
    # despite the underlying input being type="hidden"
    input_type = "hidden"
    is_hidden = False

    def __init__(self, model, **kwargs):
        self.target_model = model

        model_name = self.target_model._meta.verbose_name
        self.choose_one_text = _("Choose %s") % model_name
        self.choose_another_text = _("Choose another %s") % model_name
        self.link_to_chosen_text = _("Edit this %s") % model_name
        self.clear_choice_text = _("Clear choice")
        self.show_edit_link = True
        self.show_clear_link = True

        super().__init__(**kwargs)

    def get_value_data(self, value):
        # Given a data value (which may be None, a model instance, or a PK here),
        # extract the necessary data for rendering the widget with that value.
        # In the case of StreamField (in Wagtail >=2.13), this data will be serialised via
        # telepath https://wagtail.github.io/telepath/ to be rendered client-side, which means it
        # cannot include model instances. Instead, we return the raw values used in rendering -
        # namely: pk, display_markup and edit_url

        if not str(value).strip(): # value might be "" (Wagtail 5.0+)
            value = None

        if value is None or isinstance(value, self.target_model):
            instance = value
        else:  # assume this is an instance ID
            instance = self.target_model.objects.get(pk=value)

        app_label = self.target_model._meta.app_label
        model_name = self.target_model._meta.model_name
        model = registry.get_model(app_label, model_name)
        instance_selector = registry.get_instance_selector(model)
        display_markup = instance_selector.get_instance_display_markup(instance)
        edit_url = instance_selector.get_instance_edit_url(instance)

        return {
            "pk": instance.pk if instance else None,
            "display_markup": display_markup,
            "edit_url": edit_url,
        }

    def render_html(self, name, value, attrs):
        value_data = value
        
        original_field_html = super().render_html(name, value_data["pk"], attrs)

        app_label = self.target_model._meta.app_label
        model_name = self.target_model._meta.model_name

        embed_url = reverse(
            "wagtail_instance_selector_embed",
            kwargs={"app_label": app_label, "model_name": model_name},
        )
        # We use the input name for the embed id so that wagtail's block code will automatically
        # replace any `__prefix__` substring with a specific id for the widget instance
        embed_id = name
        embed_url += "#instance_selector_embed_id:" + embed_id

        lookup_url = reverse(
            "wagtail_instance_selector_lookup",
            kwargs={"app_label": app_label, "model_name": model_name},
        )

        return render_to_string(
            "instance_selector/instance_selector_widget.html",
            {
                "name": name,
                "is_nonempty": value_data["pk"] is not None,
                "widget": self,
                "widget_id": "%s-instance-selector-widget" % attrs["id"],
                "original_field_html": original_field_html,
                "display_markup": value_data["display_markup"],
                "edit_url": value_data["edit_url"],
            },
        )

    def get_js_config(self, id_, name):
        app_label = self.target_model._meta.app_label
        model_name = self.target_model._meta.model_name

        embed_url = reverse(
            "wagtail_instance_selector_embed",
            kwargs={"app_label": app_label, "model_name": model_name},
        )
        # We use the input name for the embed id so that wagtail's block code will automatically
        # replace any `__prefix__` substring with a specific id for the widget instance
        embed_id = name
        embed_url += "#instance_selector_embed_id:" + embed_id

        lookup_url = reverse(
            "wagtail_instance_selector_lookup",
            kwargs={"app_label": app_label, "model_name": model_name},
        )

        return {
            "input_id": id_,
            "widget_id": "%s-instance-selector-widget" % id_,
            "field_name": name,
            "embed_url": embed_url,
            "embed_id": embed_id,
            "lookup_url": lookup_url,
            "OBJECT_PK_PARAM": OBJECT_PK_PARAM,
        }

    def render_js_init(self, id_, name, value):
        config = self.get_js_config(id_, name)
        return "create_instance_selector_widget({config});".format(
            config=json.dumps(config)
        )


class InstanceSelectorAdapter(WidgetAdapter):
    js_constructor = "wagtailinstanceselector.widgets.InstanceSelector"

    def js_args(self, widget):
        return [
            widget.render_html(
                "__NAME__", widget.get_value_data(None), attrs={"id": "__ID__"}
            ),
            widget.get_js_config("__ID__", "__NAME__"),
        ]

    class Media:
        js = [
            "instance_selector/instance_selector_telepath.js",
        ]


register(InstanceSelectorAdapter(), InstanceSelectorWidget)



from django.urls import re_path
from instance_selector.views import instance_selector_embed, instance_selector_lookup


urlpatterns = [
    re_path(
        r"^instance-selector/embed/(?P<app_label>[\w-]+).(?P<model_name>[\w-]+)/$",
        instance_selector_embed,
        name="wagtail_instance_selector_embed",
    ),
    re_path(
        r"^instance-selector/lookup/(?P<app_label>[\w-]+).(?P<model_name>[\w-]+)/$",
        instance_selector_lookup,
        name="wagtail_instance_selector_lookup",
    ),
]



from django.utils.html import format_html
from django.templatetags.static import static

from instance_selector import urls

from wagtail import hooks


@hooks.register("register_admin_urls")
def register_instance_selector_urls():
    return urls.urlpatterns


@hooks.register("insert_global_admin_css")
def global_admin_css():
    return format_html(
        '<link rel="stylesheet" href="{}"><script src="{}"></script><script src="{}"></script>',
        static("instance_selector/instance_selector.css"),
        static("instance_selector/instance_selector_embed.js"),
        static("instance_selector/instance_selector_widget.js"),
    )



import inspect
from django.urls import reverse
from django.template import loader
from django.utils.safestring import mark_safe


class BaseInstanceSelector:
    def __init__(self):
        self.display_template = loader.get_template(
            self.get_instance_display_template()
        )

    def get_instance_display_markup(self, instance):
        """
        Markup representing the instance's display in a widget
        """
        markup = self.display_template.render(
            {
                "display_title": self.get_instance_display_title(instance),
                "display_image_url": self.get_instance_display_image_url(instance),
                "display_image_styles": self.get_instance_display_image_styles(
                    instance
                ),
                "edit_url": self.get_instance_edit_url(instance),
            }
        )
        return mark_safe(markup)

    def get_instance_display_title(self, instance):
        """
        A textual representation of the instance
        """
        if instance:
            return str(instance)

    def get_instance_display_image_url(self, instance):
        """
        An optional url to an image representing the instance
        """
        return None

    def get_instance_display_image_styles(self, instance):
        return {"max-width": "165px", "max-height": "165px"}

    def get_instance_display_template(self):
        return "instance_selector/instance_selector_widget_display.html"

    def get_instance_edit_url(self, instance):
        """
        The url that the instance can be edited at
        """
        raise NotImplementedError

    def get_instance_selector_url(self):
        """
        The url of a view that instances can be selected from, typically a list view
        """
        raise NotImplementedError

    def get_widget_icon(self):
        return "placeholder"


class ModelAdminInstanceSelector(BaseInstanceSelector):
    model_admin = None

    def __init__(self, model_admin=None):
        super().__init__()

        if model_admin is not None:
            self.model_admin = model_admin

        if not self.model_admin:
            raise Exception(
                "`model_admin` must be defined as an attribute on %s or passed to its __init__ method"
                % (type(self))
            )

        if inspect.isclass(self.model_admin):
            raise Exception(
                "Expected an instance of a model admin class, but received %s. "
                "You may need to instantiate your model admin before passing it to the instance selector"
                % self.model_admin
            )

    def get_instance_selector_url(self):
        index_url = self.model_admin.url_helper.index_url
        return index_url

    def get_instance_edit_url(self, instance):
        if instance:
            return self.model_admin.url_helper.get_action_url("edit", instance.pk)


class WagtailUserInstanceSelector(BaseInstanceSelector):
    def get_instance_selector_url(self):
        return reverse("wagtailusers_users:index")

    def get_instance_edit_url(self, instance):
        if instance:
            return reverse("wagtailusers_users:edit", args=[instance.pk])



from setuptools import setup
import instance_selector


install_requires = ["wagtail>=4.1"]
testing_requires = ["django_webtest"]

setup(
    name="wagtail-instance-selector",
    version=instance_selector.__version__,
    packages=["instance_selector"],
    include_package_data=True,
    description="A widget for Wagtail's admin that allows you to create and select related items",
    long_description="Documentation at https://github.com/ixc/wagtail-instance-selector",
    author="The Interaction Consortium",
    author_email="admins@interaction.net.au",
    url="https://github.com/ixc/wagtail-instance-selector",
    install_requires=install_requires,
    extras_require={"testing": testing_requires},
    classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Environment :: Web Environment",
            "Framework :: Django",
            "Framework :: Django :: 3.2",
            "Framework :: Django :: 4.1",
            "Framework :: Django :: 4.2",
            "Framework :: Wagtail",
            "Framework :: Wagtail :: 4",
            "Framework :: Wagtail :: 5",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Topic :: Utilities",
        ],
)
